import re
from pathlib import Path
from vyper.compiler import CompilerData
from vyper.semantics.types.function import FunctionVisibility

def convert_type(vyper_type: str) -> str:
    vyper_type = str(vyper_type)
    if "[" in vyper_type:
        base = re.search(r'[A-Za-z]+', vyper_type).group()
        size = re.search(r'\d+', vyper_type).group()
        if base in ["String", "Bytes"]:
            return f"({base.lower()} {size})"
        else:
            return f"(array {base.lower()} {size})"
    return f":{vyper_type}"

def get_external_interface(compiler_data: CompilerData) -> str:
    interface = compiler_data.vyper_module_folded._metadata["type"]
    stem = Path(compiler_data.contract_name).stem
    # capitalize words separated by '_'
    # ex: test_interface.vy -> TestInterface
    contract_name = "".join([x.capitalize() for x in stem.split("_")]) if "_" in stem else str(stem)

    out = ";; External Interface\n"
    funcs = []
    for func in interface.members.values():
        if func.visibility == FunctionVisibility.INTERNAL or func.name == "__init__":
            continue
        args = ""
        cur_type = ""
        for name, typ in func.arguments.items():
            if str(typ) != cur_type:
                args += convert_type(typ) + " "
                cur_type = str(typ)
            args += f"{name} "
        args = "[" + args[:-1] + "]" # remove trailing space
        return_type = ""
        if func.return_type is not None:
            return_type = convert_type(func.return_type)
        mutability = func.mutability.value
        func_str = f"(defn {func.name} {args} {return_type} :{mutability})"
        funcs.append(func_str)
    body = "\n  ".join(funcs)
    out = f"{out}(definterface {contract_name}\n  {body})"
    return out



import ast as py_ast

from dasy.parser.macros import handle_macro, is_macro


import hy
import vyper.ast.nodes as vy_nodes
from hy import models

from .builtins import parse_builtin
from .core import parse_augop, parse_call
from .ops import BIN_FUNCS, BOOL_OPS, COMP_FUNCS, UNARY_OPS, is_op, parse_op
from .utils import next_nodeid, add_src_map

# namespaces with expression handlers
from . import nodes, core, macros
from dasy.builtin import functions

BUILTIN_FUNCS = BIN_FUNCS + COMP_FUNCS + UNARY_OPS + BOOL_OPS + ["in", "notin"]

NAME_CONSTS = ["True", "False"]

CONSTS = {}

ALIASES = {
    ".": "attribute",
    "quote": "tuple",
    "array": "subscript",
    "defvar": "annassign",
    "setv": "assign",
}

SRC = ""


def parse_expr(expr):

    cmd_str = ALIASES.get(str(expr[0]), str(expr[0]))

    if is_op(cmd_str):
        return parse_op(expr, cmd_str)

    if cmd_str in nodes.handlers:
        return nodes.handlers[cmd_str](expr)

    node_fn = f"parse_{cmd_str}"

    for ns in [nodes, core, macros, functions]:
        if hasattr(ns, node_fn):
            return getattr(ns, node_fn)(expr)

    if is_macro(cmd_str):
        return handle_macro(expr)

    if cmd_str.startswith("."):
        inner_node = models.Expression((models.Symbol("."), expr[1], (cmd_str[1:])))
        outer_node = models.Expression((inner_node, *expr[2:]))
        return parse_node(outer_node)

    match cmd_str:
        case "defconst":
            CONSTS[str(expr[1])] = expr[2]
            return None
        case "defimmutable" | "defimm":
            CONSTS[str(expr[1])] = None
            return None
        case "+=" | "-=" | "*=" | "/=":
            return parse_augop(expr)
        case _:
            return parse_call(expr)


def parse_node(node):
    ast_node = None

    # dispatch on type of parsed model
    match node:
        case models.Expression(node):
            ast_node = parse_expr(node)
        case models.Integer(node):
            ast_node = vy_nodes.Int(
                value=int(node), node_id=next_nodeid(), ast_type="Int"
            )
        case models.Float(node):
            raise Exception("Floating point not supported (yet)")
            # value_node = vy_nodes.Decimal(value=Decimal(float(node)), node_id=next_nodeid(), ast_type='Decimal')
            # ast_node = value_node
        case models.String(node):
            ast_node = vy_nodes.Str(
                value=str(node), node_id=next_nodeid(), ast_type="Str"
            )
        case models.Symbol(node):
            str_node = str(node)
            if str_node in CONSTS:
                ast_node = parse_node(CONSTS[str(node)])
            elif str_node in BUILTIN_FUNCS:
                ast_node = parse_builtin(node)
            elif str_node in NAME_CONSTS:
                ast_node = vy_nodes.NameConstant(
                    value=py_ast.literal_eval(str(node)),
                    node_id=next_nodeid(),
                    ast_type="NameConstant",
                )
            elif str_node.startswith("0x"):
                ast_node = vy_nodes.Hex(
                    id=next_nodeid(), ast_type="Hex", value=str_node
                )
            elif "/" in str_node:
                target, attr = str(node).split("/")
                replacement_node = models.Expression(
                    (models.Symbol("."), models.Symbol(target), models.Symbol(attr))
                )
                ast_node = parse_node(replacement_node)
            else:
                name_node = vy_nodes.Name(
                    id=str(node), node_id=next_nodeid(), ast_type="Name"
                )
                ast_node = name_node
        case models.Keyword(node):
            name_node = vy_nodes.Name(
                id=str(node), node_id=next_nodeid(), ast_type="Name"
            )
            ast_node = name_node
        case models.Bytes(byt):
            bytes_node = vy_nodes.Bytes(
                node_id=next_nodeid(), ast_type="Byte", value=byt
            )
            ast_node = bytes_node
        case models.List(lst):
            list_node = vy_nodes.List(
                node_id=next_nodeid(), ast_type="List", elements=[]
            )
            for elmt in lst:
                node = parse_node(elmt)
                list_node._children.add(node)
                list_node.elements.append(node)
                node._parent = list_node
            ast_node = list_node
        case None:
            ast_node = None
        case _:
            raise Exception(f"No match for node {node}")
    return add_src_map(SRC, node, ast_node)


def parse_src(src: str):
    global SRC
    SRC = src
    mod_node = vy_nodes.Module(
        body=[], name="", doc_string="", ast_type="Module", node_id=next_nodeid()
    )

    vars = []
    fs = []
    for element in hy.read_many(src):
        # parse each top-level form
        ast = parse_node(element)

        if isinstance(ast, list):
            for v in ast:
                v.full_source_code = src
        elif ast is not None:
            ast.full_source_code = src
            ast.lineno = element.start_line
            ast.end_lineno = element.end_line
            ast.col_offset = element.start_column
            ast.end_col_offset = element.end_column

        match ast:
            case vy_nodes.Module:
                mod_node = ast
            case vy_nodes.VariableDecl() | vy_nodes.StructDef() | vy_nodes.EventDef() | vy_nodes.InterfaceDef():
                vars.append(ast)
            case vy_nodes.FunctionDef():
                fs.append(ast)
            case list():
                for v in ast:
                    vars.append(v)
            case vy_nodes.AnnAssign():
                # top-level AnnAssign nodes should be replaced with a VariableDecl
                is_public = False
                is_immutable = False
                is_constant = False
                if isinstance(ast.annotation, vy_nodes.Call):
                    match ast.annotation.func:
                        case "public":
                            is_public = True
                        case "immutable":
                            is_immutable = True
                        case "constant":
                            is_constant = True
                new_node = vy_nodes.VariableDecl(
                    ast_type="VariableDecl",
                    node_id=next_nodeid(),
                    target=ast.target,
                    annotation=ast.annotation,
                    value=ast.value,
                    is_constant=is_constant,
                    is_public=is_public,
                    is_immutable=is_immutable,
                )
                for child in ast.get_children():
                    new_node._children.add(child)
                    child._parent = new_node
                vars.append(new_node)
            case None:
                pass
            case _:
                raise Exception(f"Unrecognized top-level form {element} {ast}")

    for e in vars + fs:
        mod_node.add_to_body(e)
        mod_node._children.add(e)
        e._parent = mod_node

    return mod_node



from hy import models


def counter_gen():
    _counter = 0
    while True:
        yield _counter
        _counter += 1


def next_node_id_maker():
    counter = counter_gen()

    def next_num():
        return next(counter)

    return next_num


def pairwise(iterable):
    "s -> (s0, s1), (s2, s3), (s4, s5), ..."
    a = iter(iterable)
    return zip(a, a)


def has_return(tree):
    match tree:
        case models.Symbol(sym) if str(sym) == "return":
            return True
        case models.Sequence(seq):
            for el in seq:
                if has_return(el):
                    return True
        case _:
            return False
    return False


next_nodeid = next_node_id_maker()


def add_src_map(src_code, element, ast_node):
    if ast_node is None:
        return None
    if isinstance(ast_node, list):
        for n in ast_node:
            n.full_source_code = src_code
            n.lineno = element.start_line
            n.end_lineno = element.end_line
            n.col_offset = element.start_column
            n.end_col_offset = element.end_column
    else:
        ast_node.full_source_code = src_code
        if hasattr(element, "start_line"):
            ast_node.lineno = element.start_line
            ast_node.end_lineno = element.end_line
            ast_node.col_offset = element.start_column
            ast_node.end_col_offset = element.end_column
    return ast_node

def filename_to_contract_name(fname: str) -> str:
    return "".join([x.capitalize() for x in fname.split("/")[-1].split(".")[0].split("_")])



# for handling venom
from vyper.ir.s_expressions import parse_s_exp
from vyper.codegen.ir_node import IRnode
from vyper.builtin_functions import STMT_DISPATCH_TABLE, BuiltinFunction
from dasy.parser.utils import next_nodeid
from vyper.compiler import phases


def parse_venom(expr):
    ir = IRnode.from_list((parse_s_exp(expr[1]))[0])
    # generate some vyper code to patch in.
    IDENTIFIER = f"__DASY_VENOM_BUILTIN_{next_nodeid()}__"
    insert_code = f"{IDENTIFIER}()"

    # dynamically generate a class on the fly
    class generated_builtin(BuiltinFunction):
        _id = IDENTIFIER
        _inputs = ()
        _return_type = None

        def build_IR(self, expr, context):
            return ir

    STMT_DISPATCH_TABLE[IDENTIFIER] = generated_builtin()

    return phases.generate_ast(insert_code, 0, "").body[0]


def parse_vyper(expr):
    return phases.generate_ast(str(expr[1]), 0, "").body[0]



import hy
from hy import read
from dasy.compiler import compile, compile_file
from dasy.main import main
from dasy.parser.output import get_external_interface
from .parser import parse
from .parser.parse import parse_src, parse_node

__version__ = "0.1.20"



from vyper.ast import nodes as vy_nodes
from vyper.ast.nodes import (
    Break,
    Pass,
    Continue,
    Log,
    Raise,
    Return,
    AugAssign,
    Assign,
    Assert,
    AnnAssign,
    Index,
)
from hy import models
from dasy import parser
import dasy
from .core import parse_annassign, process_body
from .utils import next_nodeid

handlers = {}


def get_node(node_class, *args, **kwargs):
    args_dict = kwargs
    if len(args) != 0:
        for slot, value in zip(node_class.__slots__, args):
            args_dict[slot] = value
        for slot in node_class.__slots__[len(args) :]:
            args_dict[slot] = None
        args = []
    new_node = node_class(
        ast_type=node_class.__name__, node_id=next_nodeid(), *args, **args_dict
    )
    set_parent_children(new_node, args_dict.values())
    return new_node


def set_parent_children(parent, children):
    for n in children:
        if n is not None:
            if isinstance(n, list):
                set_parent_children(parent, n)
            else:
                parent._children.add(n)
                n._parent = parent


for node_type in [Break, Pass, Continue, Log, Raise, Return, AugAssign, Assert, Index]:
    k = node_type.__name__.lower()
    handlers[k] = lambda expr, node_type=node_type: get_node(
        node_type, *[parser.parse_node(arg) for arg in expr[1:]]
    )


def parse_for(expr):
    # (for [x xs] (.append self/nums x))
    # (for [target iter] *body)
    target, iter_ = expr[1]
    target_node = parser.parse_node(target)
    iter_node = parser.parse_node(iter_)
    body_nodes = [parser.parse_node(b) for b in expr[2:]]
    body = process_body(body_nodes)
    for_node = get_node(vy_nodes.For, body=body, iter=iter_node, target=target_node)
    set_parent_children(for_node, body + [target_node, iter_node])
    return for_node


def parse_if(expr):
    # used for base case in cond expansion
    if expr[1] == models.Keyword("else"):
        if expr[3] == models.Symbol("None"):
            return parser.parse_node(expr[2])

    body_nodes = [parser.parse_node(expr[2])]
    body = process_body(body_nodes)
    else_nodes = [parser.parse_node(expr[3])] if len(expr) == 4 else []
    else_ = process_body(else_nodes)
    test = parser.parse_node(expr[1])
    if_node = get_node(vy_nodes.If, test=test, body=body, orelse=else_)
    set_parent_children(if_node, body + else_ + [test])
    return if_node


def parse_assign(expr):
    # needs some slight massaging due to the way targets/target is treated
    # the Assign class has a target slot, but it uses the first value in the
    # targets arg to fill it instead of using the target kwarg
    args = [parser.parse_node(arg) for arg in expr[1:]]
    return get_node(vy_nodes.Assign, *args, targets=[args[0]])
